/*
 * backup.c
 *
 * Routines to perform that actual backup and restore of the device.
 *
 * Copyright (C) 2004 Daniel Thompson <rfbackup@redfelineninja.org.uk> 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "rftidy.h"

#include <stdarg.h>

typedef struct TdyRecordSummary {
	UInt16 num;
	UInt32 crc;
} TdyRecordSummaryType, *TdyRecordSummaryPtr;

/* TODO: this is not an efficient implementation! */
static UInt32 crc32(unsigned char *message, int len)
{
	int i, j;
	UInt32 byte, crc, mask;

	crc = 0xFFFFFFFF;
	
	for (i=0; i<len; i++) {
		byte = message[i]; // Get next byte.
		crc = crc ^ byte;
		
		for (j = 7; j >= 0; j--) { // Do eight times.
			mask = -(crc & 1);
			crc = (crc >> 1) ^ (0xEDB88320 & mask);
		}
	}
	
	return ~crc;
}

static Handle TdyGetChecksums(DmOpenRef db, BakCallback *cb)
{
	UInt16 i, j, len;
	Handle hChecksums, hRecord;
	TdyRecordSummaryPtr pChecksums;
	
	len = DmNumRecordsInCategory(db, dmAllCategories);
	cb(0, "Stage 1: Scanning records", 0, len);
	
	hChecksums = MemHandleNew(len * sizeof(TdyRecordSummaryType));
	if (0 == hChecksums) {
		return 0;
	}
	pChecksums = MemHandleLock(hChecksums);

	for (i=0, j=0; 0 != (hRecord = DmQueryNextInCategory(db, &j, dmAllCategories)); i++, j++) {
		UInt32 sz = MemHandleSize(hRecord);
		Char *p =  MemHandleLock(hRecord);
		cb(0, 0, i, len);
		pChecksums[i].num = j;
		pChecksums[i].crc = crc32(p, sz);
		MemHandleUnlock(hRecord);
	}
	
	MemHandleUnlock(hChecksums);
	return hChecksums;
}

static Boolean TdyCompareRecords(DmOpenRef db, UInt16 r0, UInt16 r1) 
{
	Boolean result;
	MemHandle h[2];
	void *p[2];
	UInt32 sz[2];
	
	h[0] = DmQueryRecord(db, r0);
	h[1] = DmQueryRecord(db, r1);

	sz[0] = MemHandleSize(h[0]);
	sz[1] = MemHandleSize(h[1]);
	if (sz[0] != sz[1]) {
		return false;
	}
	
	p[0] = MemHandleLock(h[0]);
	p[1] = MemHandleLock(h[1]);
	
	result = (0 == MemCmp(p[0], p[1], sz[0]));
	
	MemHandleUnlock(h[0]);
	MemHandleUnlock(h[1]);
	
	if (result) {
		Err err;
		UInt16 attr[2];
		
		(void) DmRecordInfo(db, r0, &attr[0], NULL, NULL);
		(void) DmRecordInfo(db, r1, &attr[1], NULL, NULL);
		attr[0] &= dmRecAttrCategoryMask;
		attr[1] &= dmRecAttrCategoryMask;
		
		/* check whether we need to copy r0's category to r1 */
		if (attr[0] != attr[1] && attr[1] == dmUnfiledCategory) {
			(void) DmRecordInfo(db, r1, &attr[1], NULL, NULL);
			attr[1] = (attr[1] & ~dmRecAttrCategoryMask) | dmRecAttrDirty | attr[0];
			(void) DmSetRecordInfo(db, r1, &attr[1], NULL);	
		}
		
		err = DmDeleteRecord(db, r0);
		ErrNonFatalDisplayIf(errNone != err, "Cannot delete record");
		
		err = DmMoveRecord(db, r0, DmNumRecords(db));
		ErrNonFatalDisplayIf(errNone != err, "Cannot move deleted record to back of database");
	}
	
	return result;
}

static void TdyCompareChecksums(DmOpenRef db, Handle hChecksums, BakCallback *cb)
{
	Int32 i, j;
	UInt16 len;
	TdyRecordSummaryPtr pChecksums;
	
	len = DmNumRecordsInCategory(db, dmAllCategories);	
	cb(0, "Stage 2: Finding duplicates", 0, len);
	
	pChecksums = MemHandleLock(hChecksums);
	
	for (i=(len-1); i >= 0; i--) {
		cb(0, 0, (len - i), len);
		
		for (j=i-1; j >= 0; j--) {
			if (pChecksums[i].crc == pChecksums[j].crc) {
				if (TdyCompareRecords(db, pChecksums[i].num, pChecksums[j].num)) {
					break;
				}
			}
		}
	}
	
	MemHandleUnlock(hChecksums);
}

static Err TdyIdentifyDuplicates(DmOpenRef db, BakCallback *cb)
{
	Handle hChecksums;
	
	hChecksums = TdyGetChecksums(db, cb);
	if (0 == hChecksums) {
		return dmErrMemError;
	}
	
	TdyCompareChecksums(db, hChecksums, cb);
	MemHandleFree(hChecksums);
	return errNone;
}

static Err TdyDatabase(const char *dbName, BakCallback *cb)
{
	LocalID id;
	DmOpenRef db;
	Err err;
	UInt16 len;
	
	id = DmFindDatabase(0, dbName);
	if (id == NULL) {
		/* if we cannot find the database then it doesn't need tidying! */
		return errNone;
	}
	
	db = DmOpenDatabase(0, id, dmModeReadWrite);
	if (NULL == db) {
		return DmGetLastErr();
	}
	
	len = DmNumRecordsInCategory(db, dmAllCategories);
	if (0 != len) {
		cb("Tidying Address Database", "", 0, 1);
		err = TdyIdentifyDuplicates(db, cb);
	} else {
		err = errNone;
	}
	
	DmCloseDatabase(db);
	return err;	
}

Err TdyAddressBook(BakCallback *cb)
{
	return TdyDatabase("AddressDB", cb);
}

Err TdyDateBook(BakCallback *cb)
{
	return TdyDatabase("DatebookDB", cb);
}
